@@ -21,6 +21,7 @@ import (
 	"strings"
 
 	"cuelang.org/go/cue/ast"
+	"cuelang.org/go/cue/build"
 	"cuelang.org/go/cue/errors"
 	"cuelang.org/go/cue/format"
 	"cuelang.org/go/cue/load"
@@ -38,6 +39,28 @@ After you update to a new CUE release, fix helps make the necessary changes
 to your program.
 
 Without any packages, fix applies to all files within a module.
+
+
+Experiments
+
+CUE experiments are features that are not yet part of the stable language but
+are being tested for future inclusion. Some of these may introduce backwards
+incompatible changes for which there is a cue fix. The --exp flag is used to
+change a file or package to use the new, experimental semantics. Experiments
+are enabled on a per-file basis.
+
+For example, to enable the "explicitopen" experiment for all files in a package,
+you would run:
+
+	cue fix . --exp=explicitopen
+
+For this to succeed, your current language version must support the experiment.
+If an experiment has not yet been accepted for the current version, an
+@experiment attribute is added in each affected file to mark the transition as
+complete.
+
+The special value --exp=all enables all experimental features that apply to the
+current version.
 `,
 		RunE: mkRunE(c, runFixAll),
 	}
@@ -45,6 +68,9 @@ Without any packages, fix applies to all files within a module.
 	cmd.Flags().BoolP(string(flagForce), "f", false,
 		"rewrite even when there are errors")
 
+	cmd.Flags().StringSlice("exp", nil,
+		"list of experiments to port")
+
 	return cmd
 }
 
@@ -54,10 +80,19 @@ func runFixAll(cmd *Command, args []string) error {
 		opts = append(opts, fix.Simplify())
 	}
 
+	if exps, err := cmd.Flags().GetStringSlice("exp"); err == nil && len(exps) > 0 {
+		opts = append(opts, fix.Experiments(exps...))
+	}
+
+	_, errs := fixInstances(cmd, args, flagForce.Bool(cmd), opts...)
+	return errs
+}
+
+func fixInstances(cmd *Command, args []string, force bool, opts ...fix.Option) ([]*build.Instance, errors.Error) {
 	if len(args) == 0 {
 		args = []string{"./..."}
 
-		dir := rootWorkingDir
+		dir := rootWorkingDir()
 		for {
 			if _, err := os.Stat(filepath.Join(dir, "cue.mod")); err == nil {
 				args = appendDirs(args, filepath.Join(dir, "cue.mod", "gen"))
@@ -68,27 +103,28 @@ func runFixAll(cmd *Command, args []string) error {
 
 			dir = filepath.Dir(dir)
 			if info, _ := os.Stat(dir); !info.IsDir() {
-				return errors.Newf(token.NoPos, "no module root found")
+				return nil, errors.Newf(token.NoPos, "no module root found")
 			}
 		}
 	}
 
 	instances := load.Instances(args, &load.Config{
-		Tests: true,
-		Tools: true,
+		Tests:   true,
+		Tools:   true,
+		Package: "*",
 	})
 
 	errs := fix.Instances(instances, opts...)
 
-	if errs != nil && flagForce.Bool(cmd) {
-		return errs
+	if errs != nil && !force {
+		return nil, errs
 	}
 
 	done := map[*ast.File]bool{}
 
 	for _, i := range instances {
 		for _, f := range i.Files {
-			if done[f] || !strings.HasSuffix(f.Filename, ".cue") {
+			if done[f] || (f.Filename != "-" && !strings.HasSuffix(f.Filename, ".cue")) {
 				continue
 			}
 			done[f] = true
@@ -98,14 +134,19 @@ func runFixAll(cmd *Command, args []string) error {
 				errs = errors.Append(errs, errors.Promote(err, "format"))
 			}
 
-			err = os.WriteFile(f.Filename, b, 0644)
-			if err != nil {
-				errs = errors.Append(errs, errors.Promote(err, "write"))
+			if f.Filename == "-" {
+				if _, err := cmd.OutOrStdout().Write(b); err != nil {
+					return nil, errors.Promote(err, "format")
+				}
+			} else {
+				if err := os.WriteFile(f.Filename, b, 0666); err != nil {
+					errs = errors.Append(errs, errors.Promote(err, "write"))
+				}
 			}
 		}
 	}
 
-	return errs
+	return instances, nil
 }
 
 func appendDirs(a []string, base string) []string {
