@@ -21,7 +21,6 @@ import (
 	"encoding/json"
 	"fmt"
 	"io"
-	"strings"
 
 	"cuelang.org/go/cue"
 	"cuelang.org/go/cue/ast"
@@ -238,7 +237,7 @@ func patchExpr(n ast.Node, patchPos func(n ast.Node)) {
 
 				// TODO(legacy): remove checking for '_' prefix once hidden
 				// fields are removed.
-				if !ast.IsValidIdent(u) || strings.HasPrefix(u, "_") {
+				if ast.StringLabelNeedsQuoting(u) {
 					break // keep string
 				}
 
