@@ -39,6 +39,7 @@ import (
 )
 
 func TestDefinition(t *testing.T) {
+	t.Parallel()
 	test := cuetxtar.TxTarTest{
 		Root:   "./testdata/main",
 		Name:   "definition",
@@ -55,9 +56,6 @@ func TestDefinition(t *testing.T) {
 		}
 		v.Finalize(eval.NewContext(r, v))
 
-		// TODO: do we need to evaluate v? In principle not necessary.
-		// v.Finalize(eval.NewContext(r, v))
-
 		file, errs := export.Def(r, "", v)
 		errors.Print(t, errs, nil)
 		_, _ = t.Write(formatNode(t.T, file))
@@ -77,8 +75,8 @@ func formatNode(t *testing.T, n ast.Node) []byte {
 // TestGenerated tests conversions of generated Go structs, which may be
 // different from parsed or evaluated CUE, such as having Vertex values.
 func TestGenerated(t *testing.T) {
-	cuetdtest.FullMatrix.Do(t, func(t *cuetdtest.M) {
-		ctx := t.Context()
+	cuetdtest.FullMatrix.Do(t, func(t *testing.T, m *cuetdtest.M) {
+		ctx := m.CueContext()
 
 		testCases := []struct {
 			in  func(ctx *adt.OpContext) (adt.Expr, error)
@@ -348,8 +346,8 @@ func TestFromAPI(t *testing.T) {
 	}}
 	// Issue #1204
 	for _, tc := range testCases {
-		cuetdtest.FullMatrix.Run(t, "", func(t *cuetdtest.M) {
-			ctx := t.Context()
+		cuetdtest.FullMatrix.Run(t, "", func(t *testing.T, m *cuetdtest.M) {
+			ctx := m.CueContext()
 
 			v := ctx.BuildExpr(tc.expr)
 
